#include "test/Service.h"
#include "test/Tester.h"

#include <boost/algorithm/string/join.hpp>
#include <boost/format.hpp>
#include <boost/range/adaptors.hpp>
#include <nlohmann/json.hpp>

#include <chrono>
#include <functional>
#include <thread>


namespace xs { namespace test {

namespace detail {
    std::string toString(const nlohmann::json& object)
    {
        return object.is_string() ? object.get<std::string>()
                                  : object.dump();
    }
}

Tester::Tester(const boost::filesystem::path& file)
: config_(file)
, server_(
    "xs-test", { }, { }, 0, nullptr
)
{
    registerServices();
}

void Tester::registerServices() noexcept
{
    acceptCommands_.clear();
    acceptEvents_.clear();

    using namespace std::placeholders;
    for (const auto& service : config_.getServices()) {
        if (service.isCommand()) {
            server_.onCmd(
                service.name_,
                std::bind(&Tester::onCommand, this, _1, _2)
            );
            acceptCommands_[service.name_] = &service;
        } else {
            server_.onEvent(
                service.name_,
                std::bind(&Tester::onEvent, this, _1, _2)
            );
            acceptEvents_[service.name_] = &service;
        }
    }
} 

void Tester::resetCounters() noexcept
{
    receivedCommandCount_.clear();
    receivedEventCount_.clear();
}

Tester::Report Tester::createReport(const std::string& response, const Expectation& expectation) noexcept
{
    Report report;
    if (expectation.value_.is_string()) {
         report.pass = (response == expectation.value_.get<std::string>());
    } else {
         try {
             report.pass = (nlohmann::json::parse(response) == expectation.value_);
         } catch (...) {

         }
    }

    if (!expectation.runs_.empty()) {
        for (const auto& command : expectation.runs_) {
            if (!receivedCommandCount_[command]) {
                report.missedCommands.emplace(command);
                report.pass = false;
            }
        }
    }

    if (!expectation.posts_.empty()) {
        for (const auto& event : expectation.posts_) {
            if (!receivedEventCount_[event]) {
                report.missedEvents.emplace(event);
                report.pass = false;
            }
        }
    }

    return report;
}

std::string Tester::onCommand(const std::string& name, const std::string& params) noexcept
{
    const auto found = acceptCommands_.find(name);
    const auto* service = std::get<1>(*found);
    const auto& response = service->response_;

    receivedCommandCount_[name]++;

    return detail::toString(*response);
}

void Tester::onEvent(const std::string& name, const std::string& params) noexcept
{
    const auto found = acceptEvents_.find(name);
    const auto* service = std::get<1>(*found);

    receivedEventCount_[name]++;
}

void Tester::run() noexcept
{
    for (const auto& expectation : config_.getExpectations()) {
        resetCounters();

        const auto response = server_.runCmd(expectation.name_, "");
        const auto report = createReport(response, expectation);

        if (report.pass) {
            std::cout << "[PASSED] run command: " << expectation.name_ << std::endl;
            continue;
        }

        std::cout << "[FAILED] run command: " << expectation.name_ << std::endl;
        if (report.missedCommands.empty() && report.missedEvents.empty()) {
            std::cout << boost::format(R"([REASON] unexpected response: "%1%"(real) vs %2%(expect))")
                         % response % detail::toString(expectation.value_)
                      << std::endl;
        } else if (!report.missedCommands.empty()) {
            std::cout << boost::format("[REASON] missed commands: [%1%]")
                         % boost::algorithm::join(report.missedCommands, ", ")
                      << std::endl;
        } else {
            std::cout << boost::format("[REASON] missed events: [%1%]")
                         % boost::algorithm::join(report.missedEvents, ", ")
                      << std::endl;
        }
    }
}

} }
